'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var React = require('react');
var React__default = _interopDefault(React);
var lodash = require('lodash');

function __awaiter(thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator.throw(value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
}

// XXX: Provide a default value that prints an error about
// lack of connection.

var _createContext = React.createContext(null),
    Provider = _createContext.Provider,
    Consumer = _createContext.Consumer;

var ThunderProvider = function ThunderProvider(props) {
    return React__default.createElement(Provider, { value: props.connection }, props.children);
};

function Mutation(props) {
    var child = function child(connection) {
        var runMutation = function runMutation(variables) {
            return connection.mutate({
                query: props.query,
                variables: variables
            });
        };
        return props.children(runMutation);
    };
    return React__default.createElement(Consumer, null, child);
}

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) {
  return typeof obj;
} : function (obj) {
  return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
};

var classCallCheck = function (instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
};

var createClass = function () {
  function defineProperties(target, props) {
    for (var i = 0; i < props.length; i++) {
      var descriptor = props[i];
      descriptor.enumerable = descriptor.enumerable || false;
      descriptor.configurable = true;
      if ("value" in descriptor) descriptor.writable = true;
      Object.defineProperty(target, descriptor.key, descriptor);
    }
  }

  return function (Constructor, protoProps, staticProps) {
    if (protoProps) defineProperties(Constructor.prototype, protoProps);
    if (staticProps) defineProperties(Constructor, staticProps);
    return Constructor;
  };
}();

var inherits = function (subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function, not " + typeof superClass);
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      enumerable: false,
      writable: true,
      configurable: true
    }
  });
  if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass;
};

var possibleConstructorReturn = function (self, call) {
  if (!self) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return call && (typeof call === "object" || typeof call === "function") ? call : self;
};

var slicedToArray = function () {
  function sliceIterator(arr, i) {
    var _arr = [];
    var _n = true;
    var _d = false;
    var _e = undefined;

    try {
      for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
        _arr.push(_s.value);

        if (i && _arr.length === i) break;
      }
    } catch (err) {
      _d = true;
      _e = err;
    } finally {
      try {
        if (!_n && _i["return"]) _i["return"]();
      } finally {
        if (_d) throw _e;
      }
    }

    return _arr;
  }

  return function (arr, i) {
    if (Array.isArray(arr)) {
      return arr;
    } else if (Symbol.iterator in Object(arr)) {
      return sliceIterator(arr, i);
    } else {
      throw new TypeError("Invalid attempt to destructure non-iterable instance");
    }
  };
}();

function Query(props) {
    return React__default.createElement(Consumer, null, function (connection) {
        return React__default.createElement(GraphQLRenderer, Object.assign({ connection: connection }, props));
    });
}

var GraphQLRenderer = function (_React$PureComponent) {
    inherits(GraphQLRenderer, _React$PureComponent);

    function GraphQLRenderer() {
        classCallCheck(this, GraphQLRenderer);
        return possibleConstructorReturn(this, (GraphQLRenderer.__proto__ || Object.getPrototypeOf(GraphQLRenderer)).apply(this, arguments));
    }

    createClass(GraphQLRenderer, [{
        key: "componentWillMount",
        value: function componentWillMount() {
            var _props = this.props,
                query = _props.query,
                variables = _props.variables;

            this.subscribe({ query: query, variables: variables });
        }
    }, {
        key: "componentWillReceiveProps",
        value: function componentWillReceiveProps(nextProps) {
            var query = nextProps.query,
                variables = nextProps.variables;

            if (lodash.isEqual(query, this.state.query) && lodash.isEqual(variables, this.state.variables)) {
                return;
            }
            this.subscribe({ query: query, variables: variables });
        }
    }, {
        key: "componentWillUnmount",
        value: function componentWillUnmount() {
            this.unsubscribe();
        }
    }, {
        key: "render",
        value: function render() {
            var _props2 = this.props,
                query = _props2.query,
                variables = _props2.variables;
            // If the current state is valid (subscribed/cached), we can render the data.

            var hasValidValue = this.state.state === "subscribed" || this.state.state === "cached";
            // If we are loading a new query and the previous query is the same (but with different query variables),
            // show the last value if it exists.
            var hasPreviousValue = this.state.state === "loading" && lodash.isEqual(query, this.state.query);
            // If we are rendering an error, show the last value if it exists are the query
            // and its variables are exactly the same. (i.e. a refresh of the data failed).
            var hasValueFromBeforeError = this.state.state === "error" && lodash.isEqual(query, this.state.query) && lodash.isEqual(variables, this.state.variables);
            var shouldRenderValue = hasValidValue || hasPreviousValue || hasValueFromBeforeError;
            var data = {
                value: shouldRenderValue ? this.state.value : undefined,
                state: this.state.state,
                error: this.state.error
            };
            return this.props.children(data);
        }
    }, {
        key: "onData",
        value: function onData(data, query, variables) {
            var partialState = null;
            if (data.state === "subscribed" || data.state === "cached") {
                partialState = {
                    value: data.value,
                    query: query,
                    variables: variables
                };
            }
            this.setState(Object.assign({ state: data.state, error: data.error }, partialState));
        }
    }, {
        key: "subscribe",
        value: function subscribe(_ref) {
            var _this2 = this;

            var query = _ref.query,
                variables = _ref.variables;

            this.unsubscribe();
            this.subscription = this.props.connection.subscribe({
                query: query,
                variables: variables,
                observer: function observer(data) {
                    return _this2.onData(data, query, variables);
                }
            });
            this.onData(this.subscription.data(), query, variables);
        }
    }, {
        key: "unsubscribe",
        value: function unsubscribe() {
            if (this.subscription) {
                this.subscription.close();
                this.subscription = undefined;
            }
        }
    }]);
    return GraphQLRenderer;
}(React__default.PureComponent);

function graphql(query, variables) {
    return function (Component) {
        return function (props) {
            return React__default.createElement(Query, { query: query, variables: variables(props) }, function (data) {
                return React__default.createElement(Component, Object.assign({ data: data }, props));
            });
        };
    };
}

/**
 * LRUCache implements a basic LRU uing lodash's isEqual to deeply compare keys.
 */

var LRUCache = function () {
    function LRUCache(size) {
        classCallCheck(this, LRUCache);

        this.size = size;
        this.cache = [];
    }

    createClass(LRUCache, [{
        key: "add",
        value: function add(key, value) {
            for (var i = 0; i < this.cache.length; i++) {
                if (lodash.isEqual(this.cache[i].key, key)) {
                    this.cache.splice(i, 1);
                    break;
                }
            }
            this.cache.unshift({ key: key, value: value });
            while (this.cache.length > this.size) {
                this.cache.pop();
            }
        }
    }, {
        key: "find",
        value: function find(k) {
            var _iteratorNormalCompletion = true;
            var _didIteratorError = false;
            var _iteratorError = undefined;

            try {
                for (var _iterator = this.cache[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                    var _ref = _step.value;
                    var key = _ref.key;
                    var value = _ref.value;

                    if (lodash.isEqual(key, k)) {
                        return value;
                    }
                }
            } catch (err) {
                _didIteratorError = true;
                _iteratorError = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion && _iterator.return) {
                        _iterator.return();
                    }
                } finally {
                    if (_didIteratorError) {
                        throw _iteratorError;
                    }
                }
            }

            return undefined;
        }
    }]);
    return LRUCache;
}();

/**
 * Merge combines a graphql update from the websocket with an existing value
 * into an updated value.
 */
function merge(original, update) {
    if (Array.isArray(update)) {
        if (_typeof(update[0]) === "object") {
            return Object.freeze(update[0]);
        } else {
            return update[0];
        }
    }
    if ((typeof update === "undefined" ? "undefined" : _typeof(update)) !== "object" || update === null) {
        return update;
    }
    var merged = void 0;
    if (Array.isArray(original)) {
        merged = [];
        var _iteratorNormalCompletion = true;
        var _didIteratorError = false;
        var _iteratorError = undefined;

        try {
            for (var _iterator = (update.$ || [[0, original.length]])[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                var x = _step.value;

                if (Array.isArray(x)) {
                    for (var i = x[0]; i < x[0] + x[1]; i++) {
                        merged.push(original[i]);
                    }
                } else if (merged[x] === -1) {
                    merged.push(undefined);
                } else {
                    merged.push(original[x]);
                }
            }
        } catch (err) {
            _didIteratorError = true;
            _iteratorError = err;
        } finally {
            try {
                if (!_iteratorNormalCompletion && _iterator.return) {
                    _iterator.return();
                }
            } finally {
                if (_didIteratorError) {
                    throw _iteratorError;
                }
            }
        }

        delete update.$;
        var _iteratorNormalCompletion2 = true;
        var _didIteratorError2 = false;
        var _iteratorError2 = undefined;

        try {
            for (var _iterator2 = Object.keys(update)[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
                var key = _step2.value;

                merged[Number(key)] = merge(merged[Number(key)], update[key]);
            }
        } catch (err) {
            _didIteratorError2 = true;
            _iteratorError2 = err;
        } finally {
            try {
                if (!_iteratorNormalCompletion2 && _iterator2.return) {
                    _iterator2.return();
                }
            } finally {
                if (_didIteratorError2) {
                    throw _iteratorError2;
                }
            }
        }
    } else {
        merged = (typeof original === "undefined" ? "undefined" : _typeof(original)) === "object" && original !== null ? Object.assign({}, original) : {};
        var _iteratorNormalCompletion3 = true;
        var _didIteratorError3 = false;
        var _iteratorError3 = undefined;

        try {
            for (var _iterator3 = Object.keys(update)[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
                var _key = _step3.value;

                var value = update[_key];
                if (Array.isArray(value) && value.length === 0) {
                    delete merged[_key];
                } else {
                    merged[_key] = merge(merged[_key], value);
                }
            }
        } catch (err) {
            _didIteratorError3 = true;
            _iteratorError3 = err;
        } finally {
            try {
                if (!_iteratorNormalCompletion3 && _iterator3.return) {
                    _iterator3.return();
                }
            } finally {
                if (_didIteratorError3) {
                    throw _iteratorError3;
                }
            }
        }
    }
    return Object.freeze(merged);
}

var BasicWebSocket = Object.freeze({
    CONNECTING: 0,
    OPEN: 1,
    CLOSING: 2,
    CLOSED: 3
});

// PingingWebSocket is a WebSocket you can rely on. It has timeouts for all
// operations, and performs its own pings to make sure the connection to the
// backend remains functioning.
var PingingWebSocket = function () {
    function PingingWebSocket(connectFunction, listener) {
        var options = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
        classCallCheck(this, PingingWebSocket);

        // Initialize our settings and state.
        this.connectFunction = connectFunction;
        this.listener = listener;
        this.options = Object.assign({}, PingingWebSocket.defaultOptions, options);
        this.hasShutdown = false;
    }
    // send sends a message string to the server. If the WebSocket isn't open,
    // send fails silently.


    createClass(PingingWebSocket, [{
        key: "send",
        value: function send(message) {
            if (this.socket && this.socket.readyState === WebSocket.OPEN) {
                try {
                    this.socket.send(JSON.stringify(message));
                } catch (error) {
                    console.log("pingingwebsocket: this.socket.send failed: " + error);
                }
            }
        }
        // close closes the PingingWebSocket, calling onClose on the listener if the
        // WebSocket is still open. After close returns, no more functions on the
        // listener will be called.

    }, {
        key: "close",
        value: function close() {
            this.shutdown(PingingWebSocket.reasonCloseCalled);
        }
        // readyState returns the current underlying WebSocket's readyState.

    }, {
        key: "connect",
        value: function connect() {
            return __awaiter(this, void 0, void 0, /*#__PURE__*/regeneratorRuntime.mark(function _callee() {
                var _this = this;

                var socket;
                return regeneratorRuntime.wrap(function _callee$(_context) {
                    while (1) {
                        switch (_context.prev = _context.next) {
                            case 0:
                                console.log("pingingwebsocket: connecting");
                                // Start the connection timeout.
                                this.connectTimeout = setTimeout(function () {
                                    return _this.shutdown(PingingWebSocket.reasonConnectionTimeout);
                                }, this.options.connectionTimeoutMs);
                                // Invoke the connectFunction.
                                socket = void 0;
                                _context.prev = 3;
                                _context.next = 6;
                                return this.connectFunction();

                            case 6:
                                socket = _context.sent;
                                _context.next = 13;
                                break;

                            case 9:
                                _context.prev = 9;
                                _context.t0 = _context["catch"](3);

                                this.shutdown("connectFunction: " + _context.t0);
                                return _context.abrupt("return");

                            case 13:
                                if (!this.hasShutdown) {
                                    _context.next = 16;
                                    break;
                                }

                                socket.close();
                                return _context.abrupt("return");

                            case 16:
                                // Store the socket. If we need to close it later, shutdown will do so.
                                this.socket = socket;
                                // Register socket handlers. All handlers are no-ops once we have shutdown.
                                socket.onopen = function () {
                                    if (_this.hasShutdown) {
                                        return;
                                    }
                                    console.log("pingingwebsocket: opened");
                                    // Clear the timeout now that the connection has opened.
                                    if (_this.connectTimeout) {
                                        clearTimeout(_this.connectTimeout);
                                        _this.connectTimeout = undefined;
                                    }
                                    // Start pinging. We will send the first ping after pingIntervalMs.
                                    _this.sendPingTimeout = setTimeout(function () {
                                        return _this.sendPing();
                                    }, _this.options.pingIntervalMs);
                                    // Notify the listener.
                                    _this.listener.onOpen();
                                };
                                socket.onerror = function (e) {
                                    if (_this.hasShutdown) {
                                        return;
                                    }
                                    // Shutdown on error.
                                    _this.shutdown("socket error: " + e);
                                };
                                socket.onclose = function (e) {
                                    if (_this.hasShutdown) {
                                        return;
                                    }
                                    // Shutdown on close.
                                    _this.shutdown("socket close: " + e.reason);
                                };
                                socket.onmessage = function (e) {
                                    if (_this.hasShutdown) {
                                        return;
                                    }
                                    // Parse the message.
                                    var envelope = void 0;
                                    try {
                                        envelope = JSON.parse(e.data);
                                    } catch (error) {
                                        _this.shutdown("socket message JSON.parse: " + error);
                                        return;
                                    }
                                    // If it is a ping reply, clear the timeout and schedule a new ping.
                                    if (envelope.type === "echo") {
                                        if (_this.receivePingTimeout) {
                                            clearTimeout(_this.receivePingTimeout);
                                            _this.receivePingTimeout = undefined;
                                        }
                                        if (_this.sendPingTimeout) {
                                            clearTimeout(_this.sendPingTimeout);
                                            _this.sendPingTimeout = undefined;
                                        }
                                        _this.sendPingTimeout = setTimeout(function () {
                                            return _this.sendPing();
                                        }, _this.options.pingIntervalMs);
                                    }
                                    // Notify the listener.
                                    _this.listener.onMessage(envelope);
                                };

                            case 21:
                            case "end":
                                return _context.stop();
                        }
                    }
                }, _callee, this, [[3, 9]]);
            }));
        }
        // sendPing sends a ping and sets up a timeout for the response.

    }, {
        key: "sendPing",
        value: function sendPing() {
            var _this2 = this;

            if (this.sendPingTimeout) {
                clearTimeout(this.sendPingTimeout);
                this.sendPingTimeout = undefined;
            }
            this.send({ type: "echo" });
            if (this.receivePingTimeout) {
                clearTimeout(this.receivePingTimeout);
                this.receivePingTimeout = undefined;
            }
            this.receivePingTimeout = setTimeout(function () {
                return _this2.shutdown(PingingWebSocket.reasonPingTimeout);
            }, this.options.pingTimeoutMs);
        }
        // shutdown shuts down the websocket, clears all timeouts and notifies the
        // listener.

    }, {
        key: "shutdown",
        value: function shutdown(reason) {
            // Ensure shutdown is idempotent.
            if (this.hasShutdown) {
                return;
            }
            this.hasShutdown = true;
            console.log("pingingwebsocket: shutdown: " + reason);
            // Shut down the webssocket.
            if (this.socket) {
                try {
                    this.socket.close();
                } catch (error) {
                    console.log("pingingwebsocket: this.socket.close failed: " + error);
                }
                this.socket = undefined;
            }
            // Clear all timeouts.
            if (this.connectTimeout) {
                clearTimeout(this.connectTimeout);
                this.connectTimeout = undefined;
            }
            if (this.sendPingTimeout) {
                clearTimeout(this.sendPingTimeout);
                this.sendPingTimeout = undefined;
            }
            if (this.receivePingTimeout) {
                clearTimeout(this.receivePingTimeout);
                this.receivePingTimeout = undefined;
            }
            // Notify the listener.
            this.listener.onClose(reason);
        }
    }, {
        key: "readyState",
        get: function get$$1() {
            if (this.hasShutdown) {
                return WebSocket.CLOSED;
            } else if (this.socket) {
                return this.socket.readyState;
            } else {
                // If we have no socket yet, we are still waiting for the connect
                // promise to complete.
                return WebSocket.CONNECTING;
            }
        }
    }]);
    return PingingWebSocket;
}();
// defaultOptions provides a reasonable set of default PingingWebSocket options.


PingingWebSocket.defaultOptions = Object.freeze({
    connectionTimeoutMs: 30000,
    pingIntervalMs: 30000,
    pingTimeoutMs: 30000
});
PingingWebSocket.reasonPingTimeout = "ping timeout";
PingingWebSocket.reasonConnectionTimeout = "connection timeout";
PingingWebSocket.reasonCloseCalled = "close called";

var reconnectDelayMs = 10000;

var ReconnectingWebSocket = function () {
    function ReconnectingWebSocket(connectFunction, listener) {
        var _this = this;

        classCallCheck(this, ReconnectingWebSocket);

        this.socket = undefined;
        this.reconnectHandle = undefined;
        this.hadSuccess = false;
        this.reconnect = function () {
            _this.destroySocket();
            _this.hadSuccess = false;
            _this.socket = new PingingWebSocket(_this.connectFunction, {
                onOpen: function onOpen() {
                    _this.listener.onOpen();
                },
                onMessage: function onMessage(envelope) {
                    _this.hadSuccess = true;
                    _this.listener.onMessage(envelope);
                },
                onClose: function onClose(reason) {
                    _this.destroySocket();
                    // attempt to reconnect after a reconnectDelay
                    var waitMs = _this.hadSuccess ? 0 : reconnectDelayMs;
                    _this.reconnectHandle = setTimeout(_this.reconnect, waitMs);
                    _this.listener.onClose(reason);
                }
            });
            return _this.socket.connect();
        };
        this.connectFunction = connectFunction;
        this.listener = listener;
    }

    createClass(ReconnectingWebSocket, [{
        key: "send",
        value: function send(message) {
            if (this.socket) {
                this.socket.send(message);
            }
        }
    }, {
        key: "close",
        value: function close() {
            this.destroySocket();
        }
    }, {
        key: "destroySocket",
        value: function destroySocket() {
            if (this.socket) {
                this.socket.close();
                this.socket = undefined;
            }
            if (this.reconnectHandle) {
                clearTimeout(this.reconnectHandle);
                this.reconnectHandle = undefined;
            }
        }
    }, {
        key: "state",
        get: function get$$1() {
            if (this.socket) {
                switch (this.socket.readyState) {
                    case BasicWebSocket.CONNECTING:
                        return "connecting";
                    case BasicWebSocket.OPEN:
                        return "connected";
                    default:
                        return "waiting to reconnect";
                }
            }
            if (this.reconnectHandle) {
                return "waiting to reconnect";
            } else {
                return "closed";
            }
        }
    }, {
        key: "ravenTags",
        get: function get$$1() {
            return {
                readyState: this.socket && this.socket.readyState,
                state: this.state
            };
        }
    }]);
    return ReconnectingWebSocket;
}();

var ErrorNotConnected = "not connected";
var ErrorMutationTimeout = "mutation timed out";
var ErrorConnectionClosed = "connection closed";
function dataFromSubscription(subscription) {
    return Object.freeze({
        state: subscription.state,
        value: subscription.value,
        error: subscription.error
    });
}

var GraphQLError = function (_Error) {
    inherits(GraphQLError, _Error);

    function GraphQLError(message, requestId) {
        classCallCheck(this, GraphQLError);

        var _this = possibleConstructorReturn(this, (GraphQLError.__proto__ || Object.getPrototypeOf(GraphQLError)).call(this, message));

        _this.requestId = requestId;
        // Set the prototype explicitly.
        Object.setPrototypeOf(_this, GraphQLError.prototype);
        return _this;
    }

    return GraphQLError;
}(Error);

var Connection = function () {
    function Connection(connectFunction) {
        classCallCheck(this, Connection);

        this.nextId = 0;
        this.subscriptions = new Map();
        this.mutations = new Map();
        this.past = new LRUCache(100);
        this.initialRetryDelay = 1000;
        this.maxRetryDelay = 60000;
        this.mutationTimeoutMs = 10000;
        this.socket = new ReconnectingWebSocket(connectFunction, {
            onOpen: this.handleOpen.bind(this),
            onClose: this.handleClose.bind(this),
            onMessage: this.handleMessage.bind(this)
        });
        this.socket.reconnect();
    }

    createClass(Connection, [{
        key: "makeId",
        value: function makeId() {
            return (this.nextId++).toString();
        }
    }, {
        key: "subscribe",
        value: function subscribe(_ref) {
            var _this2 = this;

            var query = _ref.query,
                variables = _ref.variables,
                observer = _ref.observer;

            var id = this.makeId();
            var cached = this.past.find({
                query: query,
                variables: variables
            });
            var subscription = {
                state: cached ? "cached" : "loading",
                retryDelay: this.initialRetryDelay,
                retryHandle: undefined,
                query: query,
                variables: variables,
                observer: observer,
                value: cached,
                error: undefined
            };
            this.subscriptions.set(id, subscription);
            if (this.socket.state === "connected") {
                this.send({
                    id: id,
                    type: "subscribe",
                    message: { query: query, variables: variables }
                });
            }
            return {
                close: function close() {
                    var sub = _this2.subscriptions.get(id);
                    if (!sub) {
                        return;
                    }
                    if (sub.value !== undefined) {
                        _this2.past.add({ query: query, variables: variables }, sub.value);
                    }
                    _this2.subscriptions.delete(id);
                    if (_this2.socket.state === "connected") {
                        _this2.send({ id: id, type: "unsubscribe" });
                    }
                },
                data: function data() {
                    return dataFromSubscription(subscription);
                }
            };
        }
    }, {
        key: "mutate",
        value: function mutate(_ref2) {
            var _this3 = this;

            var query = _ref2.query,
                variables = _ref2.variables;

            var id = this.makeId();
            if (this.socket.state === "connected") {
                this.send({ id: id, type: "mutate", message: { query: query, variables: variables } });
            } else {
                return Promise.reject(new Error(ErrorNotConnected));
            }
            return new Promise(function (resolve, reject) {
                var mutation = {
                    query: query,
                    variables: variables,
                    timeout: setTimeout(function () {
                        return mutation.reject(new Error(ErrorMutationTimeout));
                    }, _this3.mutationTimeoutMs),
                    resolve: resolve,
                    reject: reject
                };
                _this3.mutations.set(id, mutation);
            });
        }
    }, {
        key: "notify",
        value: function notify(subscription) {
            subscription.observer(dataFromSubscription(subscription));
        }
    }, {
        key: "retry",
        value: function retry(id) {
            var subscription = this.subscriptions.get(id);
            if (subscription === undefined) {
                return;
            }
            this.send({
                id: id,
                type: "subscribe",
                message: { query: subscription.query, variables: subscription.variables }
            });
        }
    }, {
        key: "open",
        value: function open() {
            this.socket.reconnect();
        }
    }, {
        key: "close",
        value: function close() {
            this.socket.close();
        }
    }, {
        key: "send",
        value: function send(message) {
            this.socket.send(message);
        }
    }, {
        key: "handleOpen",
        value: function handleOpen() {
            var _iteratorNormalCompletion = true;
            var _didIteratorError = false;
            var _iteratorError = undefined;

            try {
                for (var _iterator = this.subscriptions[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                    var _ref3 = _step.value;

                    var _ref4 = slicedToArray(_ref3, 2);

                    var id = _ref4[0];
                    var subscription = _ref4[1];

                    this.send({
                        id: id,
                        type: "subscribe",
                        message: {
                            query: subscription.query,
                            variables: subscription.variables
                        }
                    });
                    if (subscription.retryHandle) {
                        clearTimeout(subscription.retryHandle);
                        subscription.retryHandle = undefined;
                    }
                }
            } catch (err) {
                _didIteratorError = true;
                _iteratorError = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion && _iterator.return) {
                        _iterator.return();
                    }
                } finally {
                    if (_didIteratorError) {
                        throw _iteratorError;
                    }
                }
            }
        }
    }, {
        key: "handleClose",
        value: function handleClose() {
            var _iteratorNormalCompletion2 = true;
            var _didIteratorError2 = false;
            var _iteratorError2 = undefined;

            try {
                for (var _iterator2 = this.subscriptions[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
                    var _ref5 = _step2.value;

                    var _ref6 = slicedToArray(_ref5, 2);

                    var subscription = _ref6[1];

                    if (subscription.state === "subscribed") {
                        subscription.state = "cached";
                        this.notify(subscription);
                    }
                }
            } catch (err) {
                _didIteratorError2 = true;
                _iteratorError2 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion2 && _iterator2.return) {
                        _iterator2.return();
                    }
                } finally {
                    if (_didIteratorError2) {
                        throw _iteratorError2;
                    }
                }
            }

            var _iteratorNormalCompletion3 = true;
            var _didIteratorError3 = false;
            var _iteratorError3 = undefined;

            try {
                for (var _iterator3 = this.mutations[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
                    var _ref7 = _step3.value;

                    var _ref8 = slicedToArray(_ref7, 2);

                    var mutation = _ref8[1];

                    mutation.reject(new Error(ErrorConnectionClosed));
                }
            } catch (err) {
                _didIteratorError3 = true;
                _iteratorError3 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion3 && _iterator3.return) {
                        _iterator3.return();
                    }
                } finally {
                    if (_didIteratorError3) {
                        throw _iteratorError3;
                    }
                }
            }

            this.mutations.clear();
        }
    }, {
        key: "handleMessage",
        value: function handleMessage(envelope) {
            var _this4 = this;

            var subscription = void 0;
            var mutation = void 0;
            switch (envelope.type) {
                case "update":
                    subscription = this.subscriptions.get(envelope.id);
                    if (subscription !== undefined) {
                        if (subscription.state !== "subscribed") {
                            subscription.state = "subscribed";
                            subscription.error = undefined;
                            subscription.retryDelay = this.initialRetryDelay;
                        }
                        subscription.value = merge(subscription.value, envelope.message);
                        this.notify(subscription);
                    }
                    break;
                case "result":
                    mutation = this.mutations.get(envelope.id);
                    if (mutation !== undefined) {
                        mutation.resolve(merge(null, envelope.message));
                        clearTimeout(mutation.timeout);
                        this.mutations.delete(envelope.id);
                    }
                    break;
                case "error":
                    var requestId = envelope.metadata && envelope.metadata.requestId;
                    var requestIdMessage = requestId ? "RequestID:\n" + requestId + "\n" : "";
                    subscription = this.subscriptions.get(envelope.id);
                    if (subscription !== undefined) {
                        console.error("Subscription failed.", requestIdMessage, "Query:\n", subscription.query, "\nVariables:\n", subscription.variables, "\nError:\n", envelope.message);
                        subscription.state = "error";
                        subscription.error = new GraphQLError(envelope.message, requestId);
                        subscription.retryHandle = setTimeout(function () {
                            return _this4.retry(envelope.id);
                        }, subscription.retryDelay);
                        subscription.retryDelay = Math.min(this.maxRetryDelay, subscription.retryDelay * 2);
                        this.notify(subscription);
                    }
                    mutation = this.mutations.get(envelope.id);
                    if (mutation !== undefined) {
                        console.error("Mutation failed.", requestId, "Query:\n", mutation.query, "\nVariables:\n", mutation.variables, "\nError:\n", envelope.message);
                        mutation.reject(new GraphQLError(envelope.message, requestId));
                        clearTimeout(mutation.timeout);
                        this.mutations.delete(envelope.id);
                    }
                    break;
                default:
                    break;
            }
        }
    }]);
    return Connection;
}();

exports.Mutation = Mutation;
exports.Query = Query;
exports.graphql = graphql;
exports.Connection = Connection;
exports.ThunderProvider = ThunderProvider;
